(*Copyright (C) 2019, Daniel Wyckoff*)
(*This file is part of BooleanAlgebrasIntro2.

BooleanAlgebrasIntro2 is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

BooleanAlgebrasIntro2 is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BooleanAlgebrasIntro2.  If not, 
see <http://www.gnu.org/licenses/>.*)

Require Import EqDec.
Require Import NaryFunctions.
Require Import Equality.
Require Import SetUtilities.
Require Import SetUtilities1_5.
Require Import TypeUtilities.
Require Import TypeUtilities2.
Require Import ListUtilities.
Require Import FunctionalExtensionality.
Require Import LogicUtilities.
Require Import ArithUtilities.
Require Import FunctionProperties.
Require Import FunctionProperties2.
Require Import Permutations.

(*Version 0.3 bool2 -- certified -- Version 8.4 pl4 Coq.*)

(*This brief stray into function algebras
follows the notation and text of Dietlinde Lau's "Function Algbebras on 
Finite Sets" closely, just as the Boolean portion follows Steven Givant's
"Introduction to Boolean Algebras.", and the current focus, Partial
Algebras, follows Reichel's "Structural Induction on Partial Algebras." *)


(*The type of functions (operations) in the usual sense, 
of specified arity.*)
(*All types of functions are assumed to be finite.*)

Definition function (T:Type) (arity:nat) : Type :=
  T^arity -> T.

Definition function_transfer {T a b} (pf:a = b) (f:function T a) : function T b :=
  transfer_dep pf f.

Definition function_transfer_r {T a b} (pf:b = a) (f:function T a) : function T b :=
  transfer_dep_r pf f.

Lemma function_transfer_eq_refl :
  forall {T a} (f:function T a),
    function_transfer eq_refl f = f.
  unfold function_transfer; intros; rewrite transfer_dep_eq_refl; auto.
Qed.

Lemma function_transfer_r_eq_refl :
  forall {T a} (f:function T a),
    function_transfer_r eq_refl f = f.
  unfold function_transfer_r; intros; rewrite transfer_dep_r_eq_refl; auto.
Qed.

Lemma finite_function :
  forall {T} n,
    FiniteT T ->
    FiniteT (function T n).
  unfold function; intros.
  apply finite_exp; auto.
  apply finite_nprod; auto.
Qed.

Lemma function_eq_dec :
  forall {T} n,
    FiniteT T ->
    type_eq_dec (function T n).
  intros; red; intros; apply finite_eq_dec; apply finite_function; auto.
Qed.


Definition arb_function (T:Type) := {n:nat & (function T n)}.

Definition set_of_functions (T:Type) := Ensemble (arb_function T).
 
Definition all_functions (T:Type) : Ensemble (set_of_functions T) :=
  Full_set _.

  
Definition set_of_functions_n (T:Type) (n:nat) : Type :=
  Ensemble (function T n).

Definition all_functions_n (T:Type) (arity:nat) : set_of_functions_n T arity :=
  Full_set _.


Lemma finite_set_of_functions_n :
  forall {T n},
    FiniteT T ->
    forall (E:set_of_functions_n T n),
      Finite E.
  intros; apply FiniteT_Finite; apply finite_function; auto.
Qed.

Lemma finite_all_functions_n :
  forall T n,
    FiniteT T ->
    Finite (all_functions_n T n).
  intros; apply finite_set_of_functions_n; auto.
Qed.


Lemma arb_function_eq_dec :
  forall {T},
    FiniteT T ->
    type_eq_dec (arb_function T).
  intros T h0; red; intros f g. 
  destruct f as [m f], g as [n g].
  destruct (nat_eq_dec m n) as [|h1]; subst.
  pose proof (function_eq_dec n h0) as h1.
  red in h1.
  specialize (h1 f g). 
  destruct h1; subst; intuition.
  right.
  contradict h1.
  apply existT_injective1 in h1; auto.
Qed.


Definition app0 {T} (f:function T 0) : T :=
  f tt.

Definition app1 {T} (f:function T 1) (x:T) : T :=
  f (x, tt).

Definition app2 {T} (f:function T 2) (x y:T) : T :=
  f (x, (y, tt)).

Definition app3 {T} (f:function T 3) (x y z:T) : T :=
  f (x, (y, (z, tt))).

Definition app4 {T} (f:function T 4) (x1 x2 x3 x4:T) : T :=
  f (x1, (x2, (x3, (x4, tt)))).

Definition app5 {T} (f:function T 5) (x1 x2 x3 x4 x5:T) : T :=
  f (x1, (x2, (x3, (x4, (x5, tt))))).

Definition app6 {T} (f:function T 6) (x1 x2 x3 x4 x5 x6:T) : T :=
  f (x1, (x2, (x3, (x4, (x5, (x6, tt)))))).

Definition app7 {T} (f:function T 7) (x1 x2 x3 x4 x5 x6 x7:T) : T :=
  f (x1, (x2, (x3, (x4, (x5, (x6, (x7, tt))))))).

Definition app8 {T} (f:function T 8) (x1 x2 x3 x4 x5 x6 x7 x8:T) : T :=
  f (x1, (x2, (x3, (x4, (x5, (x6, (x7, (x8, tt)))))))).

Definition app9 {T} (f:function T 9) (x1 x2 x3 x4 x5 x6 x7 x8 x9:T) : T :=
  f (x1, (x2, (x3, (x4, (x5, (x6, (x7, (x8, (x9, tt))))))))).


Definition Pnk_alg (n k:nat) : Type :=
  function {m | m < k} n.
 
Definition Pk_alg (k:nat) : Type :=
  arb_function {m | m < k}.

Definition PnA_alg (n:nat) (A:Type) : Type :=
  function A n.

Definition P_A (A:Type) : Type :=
  arb_function A.

Definition E2 : Type := {n | n < 2}.

Definition to_E2 n (pf:n < 2) : E2 :=
  exist (fun c => c < 2) n pf.

Definition e2_0 : E2 := to_E2 0 (lt_0_Sn _).
Definition e2_1 : E2 := to_E2 1 (lt_n_Sn _).


Definition unary_to_function {T} (f:T->T) : function T 1 :=
  fun x => f (fst x).

Definition binary_to_function {T} (f:T->T->T) : function T 2 :=
  fun x => f (fst x) (fst (snd x)).

Definition ternary_to_function {T} (f:T->T->T->T) : function T 3 :=
  fun x => f (fst x) (fst (snd x)) (fst (snd (snd x))).

Definition unary_to_arb_function {T} (f:T->T) : arb_function T :=
  existT _ _ (unary_to_function f).

Definition binary_to_arb_function {T} (f:T->T->T) : arb_function T :=
  existT _ _ (binary_to_function f).

Definition ternary_to_arb_function {T} (f:T->T->T->T) : arb_function T :=
  existT _ _ (ternary_to_function f).




Definition function_to_arb {A n} (f:function A n) : arb_function A :=
  existT _ n f.

(*Digit or degree or arity of [arb_function]*) 
Definition ar 
           {A} (f:arb_function A) : nat := projT1 f.

Definition arbn {T} (f:arb_function T) : function T (ar f) := projT2 f.


Definition tuple_trans {T} {m n} (pf:m = n)
           (x:T^m) : T^n :=
  transfer_dep pf x.

Lemma tuple_trans_eq_refl :
  forall {T} {n} (x:T^n),
    tuple_trans eq_refl x = x.
  unfold tuple_trans; intros; rewrite transfer_dep_eq_refl; auto.
Qed.
    
  
(*Extracts the function from an [arb_function]*)
  Definition funct {A} (f:arb_function A) :
  function A (ar f) :=
  match f with
    | existT _ f => f
  end.

Definition im_arb_function {A} (f:arb_function A) : Ensemble A :=
  Im (Full_set _) (funct f).

Lemma finite_im_arb_function :
  forall {A} (f:arb_function A),
    FiniteT A ->
    Finite (im_arb_function f).
  intros A f h1.
  unfold im_arb_function.
  apply finite_image.
  rewrite Finite_FiniteT_iff.
  apply finite_nprod; auto.
Qed.

      
Definition PnA_to_PA_ens {A n} (F:Ensemble (PnA_alg n A)) :
  Ensemble (P_A A) :=
  Im F (fun f => function_to_arb f).
 
Definition Fn_alg {A:Type} (F:Ensemble (P_A A)) (n:nat)
: Ensemble (PnA_alg n A) :=
  [f:PnA_alg n A | Inn F (function_to_arb f)].

Definition P_AB {A} (B:Ensemble A) : Ensemble (P_A A) :=
  [f:P_A A | Included (im_arb_function f) B].

Definition P_kl k l : Ensemble (Pk_alg k) :=
  [f:Pk_alg k | Included (im_proj1_sig (im_arb_function f)) (seg l)].

Definition P_A' A l : Ensemble (P_A A) :=
  [f:P_A A | cardinal _ (im_arb_function f) l].

Definition P_k' n l := P_A' {m | m < n} l.

Lemma arb_function_ext :
  forall {T} (f g:arb_function T)
    (pf:ar f = ar g),
    (forall (x:T^(ar f)),
       funct f x = funct g (tuple_trans pf x)) ->
    f = g.
  unfold ar, funct; intros T f g.
  destruct f as [m f], g as [n g].
  intros h0 h1; subst; simpl in h0; subst.
  f_equal.
  apply functional_extensionality.
  intro x.
  specialize (h1 x).
  rewrite tuple_trans_eq_refl in h1; auto.
Qed.
  
Lemma function_ext :
  forall {T:Type} {m n}
         (f:function T m)
         (g:function T n) 
         (pf:m = n),
    (forall x:T^m,
       f x = g (tuple_trans pf x)) ->
    f = function_transfer_r pf g.
  intros T m n f g h1 h2; subst.
  rewrite function_transfer_r_eq_refl.
  pose proof (arb_function_ext (existT _ _ f) (existT _ _ g)) as h3.
  simpl in h3.
  specialize (h3 eq_refl h2).
  eapply existT_inv in h3.
  rewrite transfer_dep_r_eq_refl in h3; auto.

  Grab Existential Variables.
  reflexivity.
Qed.



Definition agree_except {T n} (x y:T^n) i : Prop :=
  forall j (pf:j < n),
    i <> j ->
    nth_prod j pf x = nth_prod j pf y.

Inductive essential_var
          {T n} i (pfi:i < n) (f:function T n) : Prop :=
| essential_var_intro : forall (x y:T^n),
                          agree_except x y i ->
                          f x <> f y ->
                          essential_var i pfi f.


Definition depends {T} (f:arb_function T) i (pf:i < ar f) : Prop :=
  essential_var i pf (funct f).

Definition depends' {T n} (f:function T n) i (pf:i < n) : Prop :=
  essential_var i pf f.


Definition fictitious_var {T a}
           i (f:function T a) : Prop :=
  forall (x y:T^a),
    agree_except x y i ->
    f x =
    f y.

(*This uses LEM (NNPP).*)
Lemma essential_iff :
  forall {T i n}
         (pf:i < n) (f:function T n),
    essential_var i pf f <-> ~fictitious_var i f.
  intros T i n h2 f.
  split; intro h4.
  destruct h4.
  intro h4.
  red in h4. 
  specialize (h4 x y H).
  contradiction.
  unfold fictitious_var in h4.
 
  apply NNPP. 
  contradict h4.
  intros x y h5.
  apply NNPP.
  contradict h4.
  econstructor.
  apply h5.
  assumption.
Qed.


Inductive essential {T} (f:arb_function T) : Prop :=
| essential_intro :
    forall i (pf:i < ar f),
      essential_var i pf (funct f) ->
      essential f.



(*This takes a function from a FiniteT on [option T] and 
returns a function on [T], with values of [f] being translated one to the 
left, i.e. [f (Some %)] = [f_f_o %] -- values that map to None, are 
replaced by [def]*)
Definition function_from_option
           {T n} (def:T) : 
  function (option T) n -> 0 < n -> function T n :=
  match n with
    | 0 => fun f pflt => fun _ => False_rect _ (lt_irrefl _ pflt)
    | S n' =>  fun f _ => fun x =>
                          let y := f (im_nprod (@Some T) x) in
                          match y with
                            | None => def
                            | Some y' => y'
                          end
  end.
 

(*This takes a function from a FiniteT of cardinality [S n] and 
returns a function on [n], with values of [f] being translated one to the 
left, i.e. [f (Some %)] = [a_f_f_o %] -- values that map to None, are 
replaced by [def:T]*)
Definition arb_function_from_option
           {T} (def:T)
           (f:arb_function (option T)) : 0 < ar f -> arb_function T :=
  fun pf => existT _ (ar f) (function_from_option def (funct f) pf).


(*Projection*)
Definition e {T:Type} n i (pf:i < n) : function T n :=
  fun x => nth_prod i pf x.

(*Constant*)
Definition c {T:Type} n (a:T) : function T n :=
  fun _ => a.

(*Boolean and*)
Definition e2and : function {n:nat | n < 2} 2 :=
  fun p =>
    let (x, z) := p in
    let (y, _) := z in
    let (a, pfa) := x in
    let (b, pfb) := y in
    match a with
      | 0 => exist (fun n => n < 2) 0 (lt_0_Sn _)
      | S _ =>
        match b with
          | 0 => exist (fun n => n < 2) 0 (lt_0_Sn _)
          | S _ => exist (fun n => n < 2) 1 (lt_n_Sn _)
        end
    end.

(*Boolean and*)
Definition e2or : function {n:nat | n < 2} 2 :=
  fun p =>
    let (x, z) := p in
    let (y, _) := z in
    let (a, pfa) := x in
    let (b, pfb) := y in
    match a with
      | 0 =>
        match b with
          | 0 => exist (fun n => n < 2) 0 (lt_0_Sn _)
          | S _ => exist (fun n => n < 2) 1 (lt_n_Sn _)
        end
      | S _ =>
        exist (fun n => n < 2) 1 (lt_n_Sn _)
    end.

(*Boolean exclusive disjunction (+)*)
Definition e2xor : function {n:nat | n < 2} 2 :=
  fun p =>
    let (x, z) := p in
    let (y, _) := z in
    let (a, pfa) := x in
    let (b, pfb) := y in
    match a with
      | 0 =>
        match b with
          | 0 => exist (fun n => n < 2) 0 (lt_0_Sn _)
          | S _ => exist (fun n => n < 2) 1 (lt_n_Sn _)
        end
      | S _ =>
        match b with
          | 0 => exist (fun n => n < 2) 1 (lt_n_Sn _)
          | S _ => exist (fun n => n < 2) 0 (lt_0_Sn _)
        end
    end.

(*Boolean implication*)
Definition e2impl : function {n:nat | n < 2} 2 :=
  fun p =>
    let (x, z) := p in
    let (y, _) := z in
    let (a, pfa) := x in
    let (b, pfb) := y in
    match a with
      | 0 =>
        exist (fun n => n < 2) 1 (lt_n_Sn _)
      | S _ =>
        match b with
          | 0 => exist (fun n => n < 2) 0 (lt_0_Sn _)
          | S _ => exist (fun n => n < 2) 1 (lt_n_Sn _)
        end
    end.

Definition e2iff : function {n:nat | n < 2} 2 :=
  fun p =>
    let (x, z) := p in
    let (y, _) := z in
    let (a, pfa) := x in
    let (b, pfb) := y in
    match a with
      | 0 =>
        match b with
          | 0 => exist (fun n => n < 2) 1 (lt_n_Sn _)
          | S _ => exist (fun n => n < 2) 0 (lt_0_Sn _)
        end
      | S _ =>
        match b with
          | 0 => exist (fun n => n < 2) 0 (lt_0_Sn _)
          | S _ => exist (fun n => n < 2) 1 (lt_n_Sn _)
        end
    end.

Definition e2not : function {n | n < 2} 1 :=
  fun p =>
    let (x, _) := p in
    let (a, _) := x in
    match a with
        | 0 => exist (fun n => n < 2) 1 (lt_n_Sn _)
        | S _ => exist (fun n => n < 2) 0 (lt_0_Sn _)
    end.


  
(*These [e]-Boolean algebras don't require you to write [app2 [f]]*)
Definition eand x y := app2 e2and x y.
Definition eor x y := app2 e2or x y.
Definition exor x y := app2 e2xor x y.
Definition eiff x y := app2 e2iff x y.
Definition eimpl x y := app2 e2impl x y.
Definition enot x := app1 e2not x.


(*Commutative Boolean Functions*)
Inductive comm_boolean_function : function {n:nat | n < 2} 2 -> Prop :=
  | boolean_and : comm_boolean_function e2and
  | boolean_or : comm_boolean_function e2or
  | boolean_xor : comm_boolean_function e2xor
  | boolean_iff : comm_boolean_function e2iff.

(*The 5 boolean functions in P_2*)
Inductive boolean_function : function {n:nat | n < 2} 2 -> Prop :=
  | boolean_comm : forall f, comm_boolean_function f -> boolean_function f
  | boolean_impl : boolean_function e2impl.

Lemma comm_boolean_function_assoc :
  forall (f:function {n | n < 2} 2) x y z,
    comm_boolean_function f ->
    app2 f (app2 f x y) z = app2 f x (app2 f y z).
  unfold app2; intros f x y z h1; revert x y z; simpl; destruct h1;
  intros x y z; 
  destruct x as [x], y as [y], z as [z];
  destruct x, y, z; simpl; auto.
Qed.

Lemma and_same :
  forall x,
    eand x x = x.
  unfold eand, app2; intro x; destruct x as [x]; destruct x; simpl; auto; f_equal; try apply proof_irrelevance.
  pose proof l as h1.
  apply lt_S_n in h1.
  apply lt1 in h1; subst; f_equal; apply proof_irrelevance.
Qed.

  
Lemma or_same :
  forall x,
    eor x x = x.
  unfold eor, app2; intro x; destruct x as [x]; destruct x; simpl; auto; f_equal; try apply proof_irrelevance.
  pose proof l as h1.
  apply lt_S_n in h1.
  apply lt1 in h1; subst.
  f_equal; apply proof_irrelevance.
Qed.
  
Lemma xor_same :
  forall x,
    exor x x = exist (fun c => c < 2) 0 (lt_0_Sn _).
  unfold exor, app2; intro x; destruct x as [x]; destruct x; simpl; auto; f_equal; try apply proof_irrelevance.
Qed.

Lemma iff_same :
  forall x,
    eiff x x = exist (fun c => c < 2) 1 (lt_n_Sn _).
  unfold eiff, app2; intro x; destruct x as [x]; destruct x; simpl; auto; f_equal; try apply proof_irrelevance.
Qed.
  
Lemma impl_same :
  forall x,
    eimpl x x = exist (fun c => c < 2) 1 (lt_n_Sn _).
  unfold eiff, app2; intro x; destruct x as [x]; destruct x; simpl; auto; f_equal; try apply proof_irrelevance.
Qed.

Lemma and0 :
  forall x, eand x e2_0 = e2_0.
  unfold eand, e2_0; intro x; destruct x as [x]; destruct x; simpl; auto.
Qed.

Lemma and1 :
  forall x, eand x e2_1 = x.
  unfold eand, app2, e2_1; intro x; destruct x as [x]; destruct x; simpl; auto; f_equal; try apply proof_irrelevance.
  pose proof (lt_S_n _ _ l) as h1.
  apply lt1 in h1; subst.
  f_equal; apply proof_irrelevance.
Qed.


Lemma or0 :
  forall x, eor x e2_0 = x.
  unfold eor, e2_0, app2; intro x; destruct x as [x]; destruct x; simpl; auto; f_equal; try apply proof_irrelevance.
  pose proof (lt_S_n _ _ l) as h1.
  apply lt1 in h1; subst.
  f_equal; apply proof_irrelevance.
Qed.

Lemma or1 :
  forall x, eor x e2_1 = e2_1.
  unfold eor, e2_1, app2; intro x; destruct x as [x]; destruct x; simpl; auto.
Qed.

Lemma comm_boolean_function_compat :
  forall (f:function _ _),
    comm_boolean_function f ->
    forall x y,
      app2 f x y = app2 f y x.
  unfold app2; intros f h1; destruct h1; simpl; intros x y; destruct x as [x], y as [y]; destruct x, y; auto.
Qed.

Lemma and_not :
  forall x,
    eand x (enot x) = exist (fun x => x < 2) 0 (lt_0_Sn _).
  intro x; destruct x as [x]; destruct x; auto.
Qed.

Lemma or_not :
forall x,
    eor x (enot x) = exist (fun x => x < 2) 1 (lt_n_Sn _).
  intro x; destruct x as [x]; destruct x; auto.
Qed.

Lemma not_not :
  forall x,
    enot (enot x) = x.
  unfold enot, app1; intro x; destruct x as [x]; destruct x; simpl; f_equal.
  apply proof_irrelevance.
  pose (lt_S_n _ _ l) as h1.
  apply lt1 in h1; subst.
  f_equal.
  apply proof_irrelevance.
Qed.

Lemma demorgan_and :
  forall x y,
    enot (eand x y) = eor (enot x) (enot y).
  intros x y; destruct x as [x], y as [y]; destruct x, y; simpl; auto.
Qed.

Lemma demorgan_or :
  forall x y,
    enot (eor x y) = eand (enot x) (enot y).
  intros x y; destruct x as [x], y as [y]; destruct x, y; simpl; auto.
Qed.

Lemma impl_eq :
  forall x y,
    eimpl x y = eor (enot x) y.
  intros x y; destruct x as [x], y as [y]; destruct x, y; simpl; auto.
Qed.

Lemma not_impl :
  forall x y,
    enot (eimpl x y) = eand x (enot y).
  intros x y; destruct x as [x], y as [y]; destruct x, y; simpl; auto.
Qed.

Lemma distr_over_and :
  forall x y z,
    eor x (eand y z) =
    eand (eor x y) (eor x z).
  unfold eor, eand, app2; intros x y z; destruct x as [x], y as [y], z as [z]; destruct x, y, z; simpl; auto.
Qed.

Lemma distr_over_or :
  forall x y z,
    eand x (eor y z) =
    eor (eand x y) (eand x z).
  unfold eor, eand, app2; intros x y z; destruct x as [x], y as [y], z as [z]; destruct x, y, z; simpl; auto.
Qed.

Lemma absorb_over_and : 
  forall x y,
    eor x (eand x y) = x.
  unfold eor, eand, app2; intros x y; destruct x as [x], y as [y]; destruct x, y; simpl; auto; f_equal; try apply proof_irrelevance.
  pose proof (lt_S_n _ _ l) as h1.
  apply lt1 in h1; subst.
  f_equal; apply proof_irrelevance.
  pose proof (lt_S_n _ _ l) as h1.
  pose proof (lt_S_n _ _ l0) as h2.
  apply lt1 in h1.
  apply lt1 in h2.
  subst.
  f_equal.
  apply proof_irrelevance.
Qed.

Lemma abosrb_over_or :
  forall x y,
    eand x (eor x y) = x.
  unfold eor, eand, app2; intros x y; destruct x as [x], y as [y]; destruct x, y; simpl; auto; f_equal; try apply proof_irrelevance.
  pose proof (lt_S_n _ _ l) as h1.
  apply lt1 in h1; subst.
  f_equal; apply proof_irrelevance.
  pose proof (lt_S_n _ _ l) as h1.
  pose proof (lt_S_n _ _ l0) as h2.
  apply lt1 in h1.
  apply lt1 in h2.
  subst.
  f_equal.
  apply proof_irrelevance.
Qed.

Section OrderOnFunctions.
  Definition lt_function {T n} (pfft:FiniteT T)
             (ht:T -> (FiniteT_card T pfft) @)
             (hf:T^n ->
                 (FiniteT_card _ (finite_nprod _ n pfft)) @)
             (pffit:FiniteT_index pfft ht)
             (pffif:FiniteT_index (finite_nprod _ n pfft) hf)
  : Relation (function T n) :=
    lt_fun (finite_nprod _ n pfft) hf pfft ht pffif pffit.
End OrderOnFunctions.


        

Section SuperpositionOperations.
  
  (*The unassumed premise that 
    "[r < n] and t is surjective"
    and "[r > n] or [r = n]" and [t] is injective distinguishes edelta and 
    enabla (resp.).*)
  (*The [s] prefix stands for "superposition"*)
  Definition sdelta
             {T} {r n} {t:seg_fun n nat}
             (pft:seg_fun_endo t r) 
             (f:function T n)
  : function T r :=
    fun x => f (nprod_identify pft x).

  Definition snabla {T} {u n} {q:seg_fun n nat}
             (pfq:seg_fun_endo q u) 
             (f:function T n)
  : function T u := sdelta pfq f.

  Definition spi {T} {n} {s:seg_fun n nat}
             (pfs:seg_fun_endo s n)
             (f:function T n)
  : function T n := sdelta pfs f.


  Definition sstar
             {T} {m n i}
             (pf0:0 < i)
             (pfi:i < n) 
             (f:function T n)
             (g:function T m) :
    function T (m + (pred n)) :=
    fun (x:T^(m + (pred n))) =>
      let tri := star_split x pf0 pfi in
      let fs := fst tri in
      let s := fst (snd tri) in
      let t := snd (snd tri) in
      let gx := g s in
      let pfe := S_pred_plus_minus _ _ pf0 pfi in
      let x' := transfer_nprod
                  pfe
                  (glue_nprod (glue_nprod_elt fs gx) t) in
      f x'.
     
  
  Inductive sup_op1 {T} : forall {m n},
                            (function T m -> function T n) -> Prop :=
  | pi_intro : forall {n} (s:seg_fun n nat) (pfs:perm_seg s),
                 let pfe := inj_seg_fun_endo_endo _ pfs in
                 sup_op1 (spi pfe)
  | delta_intro : forall {r n},
                    r < n ->
                    forall (t:seg_fun n nat)
                           (pfs:surj_seg_fun_endo t r),
                      let pfe := surj_seg_fun_endo_endo _ pfs in
                      sup_op1 (sdelta pfe)
  | nabla_intro : forall {u n},
                    u > n ->
                    forall (q:seg_fun n nat)
                           (pfi:inj_seg_fun_endo q u),
                      let pfe :=
                          inj_seg_fun_endo_endo _ pfi in
                      sup_op1 (sdelta pfe).

  Inductive sup_op2 {T} : forall {m n},
                            (function T n -> function T m ->
                             function T (m + (pred n))) -> Prop :=
  | star_intro : forall {m n i} (f:function T n) (g:function T m)
                        (pf0:0 < i) (pfi:i < n),
                   sup_op2 (sstar pf0 pfi) (m := m).

  Inductive sup_op {T} : forall {U}, (U->Prop) -> Prop :=
  | sup_op1_intro : forall m n, sup_op (@sup_op1 T m n)
  | sup_op2_intro : forall m n, sup_op (@sup_op2 T m n).
  
  
End SuperpositionOperations.

Section MalTsevOperations.
  Definition mzeta {T n} (f:function T n) : function T n :=
    fun x => f (nprod_identify (seg_fun_endo_right_shift n) x).

  Definition mtau {T n} : function T n -> function T n :=
    match n with
      | 0 => fun f => f
      | S n' =>
        match n' with
          | 0 => fun f => f
          | S n'' =>
            fun f x =>
              let pf0 := lt_0_Sn (S n'') in
              let pf1 := lt_n_S _ _ (lt_0_Sn n'') in
              f (nprod_identify (transpose_seg_fun_endo pf0 pf1) x)
        end
    end.

  Definition mdelta_le {T n} (pfle:n <= 1)
             (f:function T n) : function T n :=
    f.
   
  Definition mdelta_gt {T n} (pf1:1 < n) (f:function T n)  :
    function T (pred n) :=
    fun x =>
      let pf0 := lt_S_pred _ _ pf1 in
      let pfe := eq_sym (S_pred _ _ pf1) in
      f (transfer_nprod pfe 
                        (nprod_identify (add_seg_to_identify_endo (pred n) pf0) x)).

  Definition mnabla0 {T} : function T 0 -> function T 0 := fun f => f.

  Definition mnabla {T n} (f:function T n) :
    function T (S n) :=
    fun x =>
      f (nprod_identify (tl_seg_for_fictitious_endo (S n)) x). 

  Definition mstar {T m n} (pfn:0 < n) (pfm:0 < m)
             (f:function T n) (g:function T m)
          :
    function T (pred (m + n)) :=
    fun x =>
      let pfle := le_pred_sum_pos _ _ pfm pfn in
      let pr := split_nprod x _ pfle in
      let pfe := S_pred_sum_minus _ _ pfm pfn in
      f (transfer_nprod pfe (g (fst pr), snd pr)).

  
  Inductive mal_tsev_op1 {T} :
    forall {m n}, (function T m -> function T n) -> Prop :=
  | zeta_intro : forall {n}, mal_tsev_op1 (@mzeta T n)
  | tau_intro : forall {n}, mal_tsev_op1 (@mtau T n)
  | mdelta_le_intro : forall {n} (pf:n <= 1),
                       mal_tsev_op1 (mdelta_le pf)
  | mdelta_intro : forall {n} (pf:1 < n),
                    mal_tsev_op1 (mdelta_gt pf)
  | nabla0_intro : mal_tsev_op1 mnabla0
  | mnabla_intro : forall {n}, mal_tsev_op1 (@mnabla T n).

  Inductive mal_tsev_op2 {T} :
    forall {m n}, (function T n -> function T m ->
                   function T (pred (m + n))) -> Prop :=
  | mstar_intro :
      forall {m n} (pfm:0 < m) (pfn:0 < n),
        mal_tsev_op2 (mstar pfn pfm).

  Inductive mal_tsev_op {T} : forall {U}, (U->Prop) -> Prop :=
  | mal_tsev_intro : forall m n, mal_tsev_op (@mal_tsev_op1 T m n)
  | mal_tsev_op2_intro : forall m n, mal_tsev_op (@mal_tsev_op2 T m n).


  Definition mzeta_to_arb {T} (f:arb_function T) : arb_function T :=
    existT _ _ (mzeta (projT2 f)).

  Definition mtau_to_arb {T} (f:arb_function T) : arb_function T :=
    existT _ _ (mtau (projT2 f)).


  Definition mdelta_to_arb {T} (f:arb_function T) : arb_function T.
    destruct f as [n f].
    revert f.
    induction n as [|n ih].
    intro f; refine (existT _ _ f).
    destruct n as [|n].
    intro f; refine (existT _ _ f).
    intro f.
    refine (existT _ (S n) _).
    refine (mdelta_gt (lt_1_SSn _) f).
  Defined.

  Definition mnabla_to_arb {T} (f:arb_function T) :
    arb_function T :=
    existT _ _ (mnabla (projT2 f)).

  Definition mstar_to_arb {T} (f:arb_function T) (g:arb_function T) : 
    arb_function T.
    destruct f as [n f], g as [m g].
    destruct n as [|n].
    refine (existT _ _ f).
    destruct m as [|m].
    refine (existT _ _ g).
    refine (existT _ _ (mstar (lt_0_Sn _) (lt_0_Sn _) f g)).
  Defined.
  
End MalTsevOperations.